--	=========================================================================
--	Fresnel Script FX Shader for Strata CX 5
--
--	VERSION		1.0
--	AUTHOR		Jon Bradley
--	DATE		07.25.2006
--	
--	DESCRIPTION
--	
--	The Fresnel FX shader uses the surface normal of the object with respect
--  to the camera vector to render a 'rim' light effect on objects. This is
--  also known as a Facing Ratio Shader. It can be used to add subtle shading
--  similar to a silk shader, or can be used in various blending modes
--  (Photoshop) for additional realism in your models.
--  
--  At the current time, this shader does not take into consideration any
--  level of antialiasing. Here's a compositing recommendation:
--      1. Render your scene at 2x resolution
--      2. Select your matte channel in photoshop and contract by 2-4 pixels
--      3. Apply the new matte to the fresnel layer
--      4. Downsample your image to the desired resolution and composite the
--         fresnel layer with any blending mode you wish
--  
--  USAGE
--  
--  Copy/Paste the initialization script into the upper (initialization)
--  source area of a new Script FX Shader. Copy/Paste the "SOURCE" script
--  into the source area of the same script. Apply the shader to the object
--  in your scene you wish to render a script fx shader result of.
--
--  Note: Script is not setup to be applied to multiple objects at once.
--
--  LICENSE
--  
--  This work is licensed under the Creative Commons Attribution 2.5 License.
--  http://creativecommons.org/licenses/by/2.5/legalcode
--  =========================================================================

--  BEGIN INITIALIZATION SCRIPT

local implemented = { prepare=true, pixel=true, final=true, cleanup=true }

local fresnelLayer = nil        --  Our resulting color layer
local camViewDir = nil          --  3D camera directional vector
local fresnelPixel = nil        --  Storage for our calculated grayscales
local useCount = 0              --  Count so we can release memory when completed

--  Optional: composite the results from our facing ratio calculations into
--  the final color image
local compositeFresnel = false

--  CompositeNormalLayer is a function that takes any colored pixel in
--	the rendered normal layer and copies it into the color layer.
local function CompositeFresnelLayer(l)
    local fresnelValues = l
    return function (h, v, layer, existing, result)
        fresnelValues:GetRGBAFloatPixel(h, v, result)
        return result
    end
end

fresnelFunction = CompositeFresnelLayer

--  END INITIALIZATION SCRIPT

--  BEGIN SOURCE SCRIPT

--  =========================================================================
--  The first phase in the rendering process is the 'prepare' phase. In this
--  step, the renderer will accumulate information on the object, such as
--  shader info, illumination and any geometric transformations.
--
--  In the prepare phase, we will be doing the following:
--  
--  1.  Storing the current camera 3d direction vector
--  2.  Initialize a render layer to store the results of our calculations.
--  3.  Set up our dynamic pixel to store our calculated color information
--  =========================================================================

if prepare then
    
    if prepare.forGeometry and prepare.forRendering then
         
        if not fresnelLayer then
            assert(useCount == 0)
            --  Create the fresnel layer to store our facing ratio values
            fresnelLayer = prepare.image:GetNamedLayer ("Fresnel Layer")
            if not fresnelLayer then
                fresnelLayer = prepare.image:AddLayer(s3d.image.RGBAFloatPixel.kType, "Fresnel Layer")
            end
            fresnelLayer:SetSize( prepare.image:GetColorLayer():GetSize() )
        end

        if not fresnelPixel then
            --  Create a dynamic pixel to store calculated pixel values
            fresnelPixel  = s3d.image.RGBAFloatPixel.CreateDynamicPixel()
        end
        
        --  Set the save flag on the layer so it is retained in the PSD file
        fresnelLayer:SetSaveFlag(true)
        
        if not camViewDir then
            --  Store our camera vector for the current time
            local atTime = s3d.database.GetActiveContext():GetCurrentTime()
            camViewDir = new s3d.vec.Dir3d()
           camViewDir = prepare.handler : GetCamera() : GetViewDirection(atTime)
        end

        useCount = useCount + 3
    end
    
    
    return prepare.forGeometry and prepare.forRendering
    
--  =========================================================================
--  Second phase of the rendering pipeline for Script FX shaders is the
--  "pixel" phase. In this phase, the renderer has already calculated the
--  value of the pixel at the current tracing location. We can grab surface
--  state information and most of the results from this process.
--
--  Here, we just need the surface normal and the camera vector to calculate
--  the facing ratio of the surface to use as the fresnel value. This value
--  is then stored in the new fresnelLayer we created in the "prepare" phase.
--  =========================================================================

elseif pixel then

    --  Grab our surface state for the object and its surface normal
    local s = pixel.state:GetSurfaceState()
    local norm = s.surfaceNormal

    --  Calculate the absolute value of the dot product of the camera direction
    --  and the surface normal.
    
    local n = s3d.vec.Dir3d(norm)
    local d = math.abs( n:DotProduct(camViewDir) )
    
    --  Calculate the grayscale representation of this value and fill the 
    --  fresnelLayer with the proper pixel value at the proper area we're
    --  sampling right now.
    
    fresnelPixel:SetRGBAFloatPixel(s3d.image.RGBAFloatPixel(1-d,1-d,1-d,1))
    fresnelLayer:Fill(pixel.info.area, fresnelPixel)

--  =========================================================================
--  In the "final" phase, the rendering is complete and we have access to any
--  of the rendered layers we (or the renderer) created. Here we just add a
--  simple option to composite the fresnel results into our color layer.
--  This is completely optional and coule be removed.
--  =========================================================================

elseif final then
    
    --  If compositeFresnel is true in our initialization, we'll just composite
    --  this layer on top of everything else. Only do this if you just want to
    --  end up with the fresnel layer, otherwise it'll override the actual
    --  full color image.
    
    if compositeFresnel then
        local color = final.image:GetColorLayer()
        s3d.image.ApplyFilter(fresnelFunction(fresnelLayer), color)
    end

--  =========================================================================
--  In the "cleanup" phase we're complete. Release any stored values created
--  in the initialization script. The renderer uses this script not only for
--  the rendering, but for the resources palette preview and the Script FX
--  editor dialog.
--
--  We must release these values so that they are calculated properly for any
--  preview or final image the rendering engine may be calculating.
--  =========================================================================

elseif cleanup then

    fresnelLayer = nil        --  Our resulting color layer
    camViewDir = nil          --  3D camera directional vector
    fresnelPixel = nil        --  Storage for our calculated color values
    
    --  We could loop and delete all layers or stored variables, decrementing
    --  the useCount each time, but we're explicitly deleting them above.
    --  So we'll just set useCount to 0.
    useCount = 0

end

--  END SOURCE SCRIPT
